/**********************************************************************\

File:		demo.c

Purpose:	This module handles demo initialization/shutdown and a
			dispatch for the graphic effects and fades.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program in a file named "GNU General Public License".
If not, write to the Free Software Foundation, 675 Mass Ave,
Cambridge, MA 02139, USA.

\**********************************************************************/

#include "demo.h"
#include "demo crash.h"
#include "graphics.h"
#include "demo graphics.h"
#include "wipe headers.h"
#include "reversed wipe headers.h"
#include "fade headers.h"
#include "reversed fade headers.h"
#include "program globals.h"
#include "prefs.h"
#include "MsgMenus.h"
#include "file management.h"

typedef pascal short (*wipeProcPtr)(GrafPtr sourceGrafPtr, GrafPtr destGrafPtr, Rect boundsRect);
typedef pascal short (*fadeProcPtr)(Rect boundsRect, Pattern *thePattern);

unsigned char	gIsReversed;
unsigned char	gWhichPict;
short				gUseTimerFlag = 1;	
short				gWipeStatus;
short				gWhichWipe;
short				gLastWipe;
wipeProcPtr		gTheWipe[NUM_WIPES+1], gTheReverseWipe[NUM_WIPES+1];
fadeProcPtr		gTheFade[NUM_WIPES+1], gTheReverseFade[NUM_WIPES+1];

void InitTheProgram(void)
{
	short				index;
	
	InitFiles();
	
	if (gIsReversed)
		ReverseAllWipes();
	
	gWhichWipe=0;
	
	(**(gTheWindowData[kMainWindow])).dispatchProc = (dispatchProcPtr) MainWindowDispatch;		/* demo graphics.c */
	
	((**(gTheWindowData[kMainWindow])).dispatchProc)((WindowDataHandle)gTheWindowData[kMainWindow],kStartup,0L);
	
	index=1;
	gTheFade[index++] = (fadeProcPtr) BoxOutFade;
	gTheFade[index++] = (fadeProcPtr) RandomFade;
	gTheFade[index++] = (fadeProcPtr) SpiralGyraFade;
	gTheFade[index++] = (fadeProcPtr) CircularFade;
	gTheFade[index++] = 0L;					/* CasteWipeRL can't be a fade */
	gTheFade[index++] = (fadeProcPtr) FourCornerFade;
	gTheFade[index++] = (fadeProcPtr) BoxInFade;
	gTheFade[index++] = 0L;					/* RippleWipe can't be a fade */
	gTheFade[index++] = 0L;					/* DissolveWipe can't easily be a fade */
	gTheFade[index++] = (fadeProcPtr) DiagonalFade;
	gTheFade[index++] = 0L;				/* CasteWipe can't be a fade */
	gTheFade[index++] = (fadeProcPtr) HilbertFade;
	gTheFade[index++] = (fadeProcPtr) SlideFade;
	gTheFade[index++] = (fadeProcPtr) SkipalineFade;
	gTheFade[index++] = (fadeProcPtr) SkipalineLRFade;
	gTheFade[index++] = (fadeProcPtr) ScissorsFade;
	gTheFade[index++] = (fadeProcPtr) RescueRaidersFade;
	gTheFade[index++] = (fadeProcPtr) FourCornerCenteredFade;
	gTheFade[index++] = (fadeProcPtr) TwoCornerFade;
	gTheFade[index++] = (fadeProcPtr) CircleOutFade;
	gTheFade[index++] = (fadeProcPtr) CircleInFade;
	gTheFade[index++] = (fadeProcPtr) CircleSerendipityFade;
	gTheFade[index++] = (fadeProcPtr) CircleBulgeFade;
	gTheFade[index++] = (fadeProcPtr) CornerCircleFade;
	gTheFade[index++] = (fadeProcPtr) QuadrantFade;
	gTheFade[index++] = (fadeProcPtr) QuadrantFade2;
	gTheFade[index++] = (fadeProcPtr) Skipaline2PassFade;
	gTheFade[index++] = (fadeProcPtr) SkipalineLR2PassFade;
	gTheFade[index++] = (fadeProcPtr) HGRFade;
	gTheFade[index++] = (fadeProcPtr) HGR2Fade;
	gTheFade[index++] = (fadeProcPtr) FullScrollLRFade;
	gTheFade[index++] = (fadeProcPtr) FullScrollUDFade;
	gTheFade[index++] = (fadeProcPtr) MrDoFade;
	gTheFade[index++] = (fadeProcPtr) MrDoOutdoneFade;
	gTheFade[index++] = (fadeProcPtr) SplitScrollUDFade;
	gTheFade[index++] = (fadeProcPtr) FourCornerScrollFade;
	gTheFade[index++] = (fadeProcPtr) HalvesScrollFade;
	gTheFade[index++] = (fadeProcPtr) PourScrollFade;
	gTheFade[index++] = (fadeProcPtr) QuadrantScrollFade;
	gTheFade[index++] = (fadeProcPtr) QuadrantScroll2Fade;
	if (index!=NUM_WIPES+1)
		Debugger();
	
	index=1;
	gTheReverseFade[index++] = (fadeProcPtr) BoxInFade;
	gTheReverseFade[index++] = (fadeProcPtr) RandomFade;
	gTheReverseFade[index++]=0L;				/* Spiral Gyra can't be easily reversed. */
	gTheReverseFade[index++]=(fadeProcPtr)CircularFadeReversed;
	gTheReverseFade[index++]=0L;				/* CasteWipeRight can't be a fade */
	gTheReverseFade[index++]=(fadeProcPtr)FourCornerFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)BoxOutFade;
	gTheReverseFade[index++]=0L;				/* RippleWipe Reversed can't be a fade */
	gTheReverseFade[index++]=0L;				/* DissolveWipe can't easily be a fade */
	gTheReverseFade[index++]=(fadeProcPtr)DiagonalFadeDownRight;
	gTheReverseFade[index++]=0L;				/* CasteWipeDown can't be a fade */
	gTheReverseFade[index++]=(fadeProcPtr)HilbertFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)SlideFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)SkipalineFade;
	gTheReverseFade[index++]=(fadeProcPtr)SkipalineLRFade;
	gTheReverseFade[index++]=(fadeProcPtr)ScissorsFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)RescueRaidersFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)FourCenteredFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)TwoCornerFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)CircleInFade;
	gTheReverseFade[index++]=(fadeProcPtr)CircleOutFade;
	gTheReverseFade[index++]=0L;				/* CircleSerendipity can't be easily reversed */
	gTheReverseFade[index++]=0L;				/* CircleBulge can't be easily reversed */
	gTheReverseFade[index++]=(fadeProcPtr)CornerCircleFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)QuadrantFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)QuadrantFade2Reversed;
	gTheReverseFade[index++]=(fadeProcPtr)Skipaline2PassFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)SkipalineLR2PassFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)HGRFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)HGR2FadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)FullScrollLeftFade;
	gTheReverseFade[index++]=(fadeProcPtr)FullScrollUpFade;
	gTheReverseFade[index++]=(fadeProcPtr)MrDoFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)MrDoOutdoneFadeReversed;
	gTheReverseFade[index++]=0L;				/* SplitScrollUD can't be reversed */
	gTheReverseFade[index++]=0L;				/* FourCornerScroll can't be reversed */
	gTheReverseFade[index++]=(fadeProcPtr)HalvesScrollFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)PourScrollFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)QuadrantScrollFadeReversed;
	gTheReverseFade[index++]=(fadeProcPtr)QuadrantScroll2FadeReversed;
	if (index!=NUM_WIPES+1)
		Debugger();
	
	index=1;
	gTheWipe[index++]=(wipeProcPtr)BoxOutWipe;
	gTheWipe[index++]=(wipeProcPtr)RandomWipe;
	gTheWipe[index++]=(wipeProcPtr)SpiralGyra;
	gTheWipe[index++]=(wipeProcPtr)CircularWipe;
	gTheWipe[index++]=(wipeProcPtr)CasteWipeRL;
	gTheWipe[index++]=(wipeProcPtr)FourCorner;
	gTheWipe[index++]=(wipeProcPtr)BoxInWipe;
	gTheWipe[index++]=(wipeProcPtr)RippleWipe;
	gTheWipe[index++]=(wipeProcPtr)DissolveWipe;
	gTheWipe[index++]=(wipeProcPtr)DiagonalWipe;
	gTheWipe[index++]=(wipeProcPtr)CasteWipe;
	gTheWipe[index++]=(wipeProcPtr)HilbertWipe;
	gTheWipe[index++]=(wipeProcPtr)SlideWipe;
	gTheWipe[index++]=(wipeProcPtr)Skipaline;
	gTheWipe[index++]=(wipeProcPtr)SkipalineLR;
	gTheWipe[index++]=(wipeProcPtr)Scissors;
	gTheWipe[index++]=(wipeProcPtr)RescueRaiders;
	gTheWipe[index++]=(wipeProcPtr)FourCornerCentered;
	gTheWipe[index++]=(wipeProcPtr)TwoCorner;
	gTheWipe[index++]=(wipeProcPtr)CircleOut;
	gTheWipe[index++]=(wipeProcPtr)CircleIn;
	gTheWipe[index++]=(wipeProcPtr)CircleSerendipity;
	gTheWipe[index++]=(wipeProcPtr)CircleBulge;
	gTheWipe[index++]=(wipeProcPtr)CornerCircle;
	gTheWipe[index++]=(wipeProcPtr)QuadrantWipe;
	gTheWipe[index++]=(wipeProcPtr)QuadrantWipe2;
	gTheWipe[index++]=(wipeProcPtr)Skipaline2Pass;
	gTheWipe[index++]=(wipeProcPtr)SkipalineLR2Pass;
	gTheWipe[index++]=(wipeProcPtr)HGR;
	gTheWipe[index++]=(wipeProcPtr)HGR2;
	gTheWipe[index++]=(wipeProcPtr)FullScrollLR;
	gTheWipe[index++]=(wipeProcPtr)FullScrollUD;
	gTheWipe[index++]=(wipeProcPtr)MrDo;
	gTheWipe[index++]=(wipeProcPtr)MrDoOutdone;
	gTheWipe[index++]=(wipeProcPtr)SplitScrollUD;
	gTheWipe[index++]=(wipeProcPtr)FourCornerScroll;
	gTheWipe[index++]=(wipeProcPtr)HalvesScroll;
	gTheWipe[index++]=(wipeProcPtr)PourScroll;
	gTheWipe[index++]=(wipeProcPtr)QuadrantScroll;
	gTheWipe[index++]=(wipeProcPtr)QuadrantScroll2;
	if (index!=NUM_WIPES+1)
		Debugger();
	
	index=1;
	gTheReverseWipe[index++]=(wipeProcPtr)BoxInWipe;
	gTheReverseWipe[index++]=(wipeProcPtr)RandomWipe;
	gTheReverseWipe[index++]=0L;			/* Spiral Gyra can't be easily reversed */
	gTheReverseWipe[index++]=(wipeProcPtr)CircularWipeReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)CasteWipeRight;
	gTheReverseWipe[index++]=(wipeProcPtr)FourCornerReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)BoxOutWipe;
	gTheReverseWipe[index++]=(wipeProcPtr)RippleWipeReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)DissolveWipe;
	gTheReverseWipe[index++]=(wipeProcPtr)DiagonalWipeDownRight;
	gTheReverseWipe[index++]=(wipeProcPtr)CasteWipeDown;
	gTheReverseWipe[index++]=(wipeProcPtr)HilbertWipeReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)SlideWipeReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)Skipaline;
	gTheReverseWipe[index++]=(wipeProcPtr)SkipalineLR;
	gTheReverseWipe[index++]=(wipeProcPtr)ScissorsReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)RescueRaidersReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)FourCornerCenteredReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)TwoCornerReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)CircleIn;
	gTheReverseWipe[index++]=(wipeProcPtr)CircleOut;
	gTheReverseWipe[index++]=0L;			/* CircleSerendipity can't be easily reversed */
	gTheReverseWipe[index++]=0L;			/* CircleBulge can't be easily reversed */
	gTheReverseWipe[index++]=(wipeProcPtr)CornerCircleReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)QuadrantWipeReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)QuadrantWipe2Reversed;
	gTheReverseWipe[index++]=(wipeProcPtr)Skipaline2PassReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)SkipalineLR2PassReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)HGRReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)HGR2Reversed;
	gTheReverseWipe[index++]=(wipeProcPtr)FullScrollLeft;
	gTheReverseWipe[index++]=(wipeProcPtr)FullScrollUp;
	gTheReverseWipe[index++]=(wipeProcPtr)MrDoReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)MrDoOutdoneReversed;
	gTheReverseWipe[index++]=0L;			/* SplitScrollUD can't be reversed */
	gTheReverseWipe[index++]=0L;			/* FourCornerScroll can't be reversed */
	gTheReverseWipe[index++]=(wipeProcPtr)HalvesScrollReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)PourScrollReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)QuadrantScrollReversed;
	gTheReverseWipe[index++]=(wipeProcPtr)QuadrantScroll2Reversed;
	if (index!=NUM_WIPES+1)
		Debugger();
}

void ShutDownTheProgram(void)
{
	((**(gTheWindowData[kMainWindow])).dispatchProc)((WindowDataHandle)gTheWindowData[kMainWindow], kShutdown, 0L);
	SaveThePrefs();
}

void DoFullScreenFade(short whichWipe)
{
	Rect			theRect;
	short				oldMenuBarHeight;

	ObscureCursor();
	oldMenuBarHeight=GetMBarHeight();
	
//	MBarHeight=0;
	LMSetMBarHeight(0);
	
	DrawMenuBar();
//	SetPort(WMgrPort);
	SetPort(LMGetWMgrPort());
	theRect=LMGetWMgrPort()->portRect;
	
	gWhichWipe=whichWipe;
	FadeDispatch(theRect);
	RestoreScreen(oldMenuBarHeight);	
	gLastWipe=gWhichWipe;
	gWhichWipe=0;
}

void FadeDispatch(Rect theRect)
{
	if (gIsReversed)
	{
		if (gTheReverseFade[gWhichWipe]!=0L)
			gTheReverseFade[gWhichWipe](theRect,&qd.black);
	}
	else
	{
		if (gTheFade[gWhichWipe]!=0L)
			gTheFade[gWhichWipe](theRect,&qd.black);
	}
}

void WipeDispatch(GrafPtr sourceGrafPtr, GrafPtr destGrafPtr, Rect theRect)
{
	ObscureCursor();

	gWhichPict = !gWhichPict;
	
	if ((gWipeStatus==kFadesPlusEffects) || (gWipeStatus==kFadesPlusReverseEffects))
		FadeDispatch(theRect);
	if (gWipeStatus==kFadesPlusReverseEffects)
		gIsReversed=!gIsReversed;
	
	if (gIsReversed)
	{
		if (gTheReverseWipe[gWhichWipe]!=0L)
			gTheReverseWipe[gWhichWipe](sourceGrafPtr, destGrafPtr, theRect);
		else
			CopyBits(&(sourceGrafPtr->portBits), &(destGrafPtr->portBits), &theRect, &theRect, 0, 0L);
	}
	else
	{
		if (gTheWipe[gWhichWipe]!=0L)
			gTheWipe[gWhichWipe](sourceGrafPtr, destGrafPtr, theRect);
		else
			CopyBits(&(sourceGrafPtr->portBits), &(destGrafPtr->portBits), &theRect, &theRect, 0, 0L);
	}
	
	if (gWipeStatus==kFadesPlusReverseEffects)
		gIsReversed=!gIsReversed;
	
	gLastWipe=gWhichWipe;
	gWhichWipe=0;
}
