; **************************************************************************************************
; this code taken from Inside Macintosh: Networking, chapter 7 (DDP)
; checksum code taken from Inside Macintosh Volume II
; **************************************************************************************************

; --------------------------------------------------------------------------------------------------
			INCLUDE 'APPLETALK.A'
			INCLUDE 'TRAPS.A'
			INCLUDE 'SYSERR.A'

; --------------------------------------------------------------------------------------------------
SL_Locals		PROC
			ENTRY 		free_queue, used_queue, current_qelem
				
free_queue		DC.L		0			; pointer to freeQ QHdr
used_queue		DC.L		0			; pointer to usedQ QHdr
current_qelem		DC.L		0			; pointer to current PacketBuffer record

			ENDP
				
; --------------------------------------------------------------------------------------------------
PacketBuffer		RECORD		0
qLink			DS.L		1
qType			DS.W		1
buffer_Type		DS.W		1			; DDP protocol type
buffer_NodeID		DS.W		1			; destination node
buffer_Address		DS.L		1			; source address in AddrBlock format
buffer_Hops		DS.W		1			; hop count
buffer_ActCount		DS.W		1			; length of DDP datagram
buffer_CheckSum		DS.W		1			; checksum error returned here (cksumErr
								;  or noErr)

buffer_Data		DS.B		ddpMaxData

			ENDR
					
; --------------------------------------------------------------------------------------------------
QHdr			RECORD		0
qFlags			DS.W		1
qHead			DS.L		1
qTail			DS.L		1
			ENDR

; --------------------------------------------------------------------------------------------------
; Pascal: Function SL_InitSktListener(freeQ, usedQ: QHdrPtr) : OSErr;
; C:      OSErr SL_InitSktListener( QHdrPtr freeQ, QHdrPtr usedQ );
;
SL_InitSktListener	PROC		EXPORT

StackFrame		RECORD		{A6Link},DECR		; build a stack frame record
Result1			DS.W		1			; function result retuned to caller
ParamBegin		EQU		*			; start parameters after this point
freeQ			DS.L		1			; freeQ parameter
usedQ			DS.L		1			; usedQ parameter
ParamSize		EQU		ParamBegin-*		; size of all the passed parameters
RetAddr			DS.L		1			; placeholder for return address
A6Link			DS.L		1			; placeholder for A6 link
LocalSize		EQU		*			; size of all the local variables
			ENDR
			
			WITH		StackFrame,QHdr		; use these record types
			LINK		A6,#LocalSize		; allocate your local stack frame
	
; Copy the queue header pointers into our local storage for use in the listener

			LEA		used_queue,A0		; copy usedQ into used_queue
			MOVE.L		usedQ(A6),(A0)
			
			LEA		free_queue,A0		; copy freeQ into free_queue
			MOVE.L		freeQ(A6),(A0)
			
; Release the first buffer record from freeQ and set current_qelem to it

			MOVEA.L		freeQ(A6),A1		; A1 = &freeQ
			LEA		current_qelem,A0	; copy freeQ.qHead into current_qelem
			MOVE.L		qHead(A1),(A0)
			MOVEA.L		qHead(A1),A0		; A0 = freeQ.qHead
			_Dequeue
			MOVE.W		D0,Result1(A6)		; return status
			
@1			UNLK		A6			; destroy the link
			MOVEA.L		(SP)+,A0		; pull off the return address
			ADDA.L		#ParamSize,SP		; strip off all the caller's parameters
			JMP		(A0)			; return to the caller
			ENDP

; --------------------------------------------------------------------------------------------------
; SL_TheListener
; Input:
;	D0 (byte) = packet's destination socket number
;	D1 (word) = number of bytes left to read in packet
;	A0 points to the bytes to checksum
;	A1 points to the bytes to checksum
;	A2 points to MPP's local variables
;	A3 points to next free byte in read-header area
;	A4 points to ReadPacket and ReadRest jump table
;
; Return:
;	D0 is modified
;	D3 (word) = accumulated checksum

SL_TheListener		PROC		EXPORT
			WITH PacketBuffer
			
; Get pointer to current PacketBuffer.
GetBuffer:
			LEA		current_qelem,A3	; get the pointer to PacketBuffer
			MOVE.L		(A3),A3
			MOVE.L		A3,D0			; if no PacketBuffer
			BEQ.S		NoBuffer		;  then ignore packet
			
; Read rest of packet into PacketBuffer.datagramData.
			MOVE.L		D1,D3			; read rest of packet
			LEA		buffer_data(A3),A3	; A3 = &buffer_data
			JSR		2(A4)			; call ReadRest
			BEQ.S		ProcessPacket		; if no error, continue
			BRA		RcvRTS			; if error, ignore the packet
			
; No buffer, ignore the packet.
NoBuffer:
			CLR		D3			; set to ignore packet (buffer size = 0)
			JSR		2(A4)			; call ReadRest
			BRA		GetNextBuffer		; no buffer available, so read next packet
								;  maybe there will be a buffer for the
								;  next packet
								
; Process the packet you just read in.
;  ReadRest has been called so registers A0-A3 and D0-D3 are free to use.
;  Use registers this way:
PktBuff			EQU		A0			; current PacketBuffer
MPPLocals		EQU		A2			; pointer to MPP's local variables
								;  (still setup from entry to
								;  socekt listener)
HopCount		EQU		D0			; gets the hop count
DatagramLength		EQU		D1			; determines the datagram length
SourceNetAddr		EQU		D2			; builds the source network address

ProcessPacket:
			LEA		current_qelem,PktBuff	; PktBuff = current_qelem
			MOVE.L		(PktBuff),PktBuff

; Do everything that's common to both long and short DDP headers first, clear buffer_Type
;  and buffer_NodeID to ensure their high bytes are 0.
			CLR.W		buffer_Type(PktBuff)	; clear buffer_Type
			CLR.W		buffer_NodeID(PktBuff)	; clear buffer_NodeID
			
; Clear SourceNetAddr to prepare to build network address
			MOVEQ		#0,SourceNetAddr	; build the network address in SourceNetAddr
			
; Get the hop count
			MOVE.W		toRHA+lapHdSz+ddpLength(MPPLocals),HopCount
								; get hop/length field
			ANDI.W		#DDPHopsMask,HopCount	; mask off the hop count bits
			LSR.W		#2,HopCount		; shift hop count into low bits of high byte
			LSR.W		#8,HopCount		; shift hop count into low byte
			MOVE.W		HopCount,buffer_Hops(PktBuff)
								; and move it into the PacketBuffer

; Get the packet length (including the DDP header).
			MOVE.W		toRHA+lapHdSz+ddpLength(MPPLocals),DatagramLength
								; get length field
			ANDI.W		#ddpLenMask,DatagramLength
								; mask off the hop count bits

; Now, find out if the DDP header is long or short
			MOVE.B		toRHA+lapType(MPPLocals),D3
								; get LAP type
			CMPI.B		#shortDDP,D3		; is this a long or short DDP header?
			BEQ.S		IsShortHdr		; skip if short DDP header

; It's a long DDP header
			MOVE.B		toRHA+lapHdSz+ddpType(MPPLocals),buffer_Type+1(PktBuff)
								; get DDP type
			MOVE.B		toRHA+LapHdSz+ddpDstNode(MPPLocals),buffer_NodeID+1(PktBuff)
								; get destination node from frame header
			MOVE.L		toRHA+LapHdSz+ddpSrcNet(MPPLocals),SourceNetAddr
								; source network in high word, source node
								;  in low byte
			LSL.W		#8,SourceNetAddr	; shift source node up to high byte
								;  of low word; get source socket from DDP header
			MOVE.B		toRHA+lapHdSz+ddpSrcSkt(MPPLocals),SourceNetAddr
			SUB.W		#ddpType+1,DatagramLength
								; DatagramLength = number of bytes in datagram
			BRA.S		MoveToBuffer
			
; Determine if there is a checksum.
			TST.W		toRHA+lapHdSz+ddpChecksum(MPPLocals)
								; does packet have checksum?
			BEQ.S		noChecksum

; Calculate checksum for the DDP header.
			MOVE.L		DatagramLength,-(SP)	; save DatagramLength (D1)
			CLR		D3			; set checksum to 0
			MOVEQ		#ddphSzLong-ddpDstNet,D1
								; D1 = length of header part to
								;  checksum pointer to destination
								;  network number in DDP header
			LEA		toRHA+lapHdSz+ddpDstNet(MPPLocals),A1

; D1 (word) = number of bytes to checksum
; D3 (word) = current checksum
; A1 points to the bytes to checksum

DoChecksum1:
			CLR.W		D0			; clear high byte
			SUBQ.W		#1,D1			; decrement count for DBRA
Loop1:			MOVE.B		(A1)+,D0		; read a byte into D0
			ADD.W		D0,D3			; accumulate checksum
			ROL.W		#1,D3			; rotate left one bit
			DBRA		D1,Loop1		; loop if more bytes

; Calculate checksum for the data portion of the packet (if any).
			MOVE.L		buffer_Data(PktBuff),A1	; pointer to datagram
			MOVE.L		(SP)+,DatagramLength	; restore DatagramLength (D1)
			MOVE.L		DatagramLength,-(SP)	; save DatagramLength (D1) before
								;  calling SL_DoChksum
			BEQ.S		TestChecksum		; don't checksum datagram if its length = 0

; D1 (word) = number of bytes to checksum
; D3 (word) = current checksum
; A1 points to the bytes to checksum

DoChecksum2:
			CLR.W		D0			; clear high byte
			SUBQ.W		#1,D1			; decrement count for DBRA
Loop2:			MOVE.B		(A1)+,D0		; read a byte into D0
			ADD.W		D0,D3			; accumulate checksum
			ROL.W		#1,D3			; rotate left one bit
			DBRA		D1,Loop2		; loop if more bytes


TestChecksum:
			MOVE.L		(SP)+,DatagramLength	; restore DatagramLength (D1)

; Now make sure the checksum is OK
			TST.W		D3			; is the calculated value 0?
			BNE.S		NotZero			; if nonzero, go and use it
			SUBQ.W		#1,D3			; if 0, make it -1
			
NotZero:
			CMP.W		toRHA+lapHdSz+ddpChecksum(MPPLocals),D3
			BNE.S		ChecksumErr		; bad checksum
			MOVE.W		#0,buffer_CheckSum(A0)	; no errors
			BRA.S		noChecksum
			
ChecksumErr:
			MOVE.W		#ckSumErr,buffer_CheckSum(PktBuff)
								; checksum error

noChecksum:
			BRA.S		MoveToBuffer

; It's a short DDP header.
IsShortHdr:
			MOVE.B		toRHA+lapHdSz+sddpType(MPPLocals),buffer_Type+1(pktBuff)
								; get DDP type
			MOVE.B		toRHA+lapDstAdr(MPPLocals),buffer_NodeID+1(pktBuff)
								; get destination node from LAP header
			MOVE.B		toRHA+lapSrcAdr(MPPLocals),SourceNetAddr
								; get source node from LAP header
			LSL.W		#8,SourceNetAddr	; shift src node up to high byte of low word
			MOVE.B		toRHA+lapHdSz+sddpSrcSkt(MPPLocals),SourceNetAddr
								; get source socket from short DDP header
			SUB.W		#sddpType+1,DatagramLength
								; DatagramLength = number of bytes in datagram

MoveToBuffer:
			MOVE.L		SourceNetAddr,buffer_Address(PktBuff)
								; move source network address into PacketBuffer
			MOVE.W		DatagramLength,buffer_ActCount(PktBuff)
								; move datagram length into PacketBuffer

; Write the packet into the used queue and get another buffer from the free queue
;  for the next packet
			LEA		used_queue,A1		; A1 = &used_queue
			MOVE.L		(A1),A1			; A1 = used_queue (pointer to usedQ)
			_Enqueue				; put the PacketBuffer in the used queue
			
GetNextBuffer:
			LEA		free_queue,A1		; A1 = &free_queue
			MOVE.L		(A1),A1			; A1 = free_queue (pointer to freeQ)
			LEA		current_qelem,A0	; copy freeQ.qHead into current_qelem
			MOVE.L		qHead(A1),(A0)
			MOVEA.L		qHead(A1),A0		; A0 = freeQ.qHead
			_Dequeue

RcvRTS:
			RTS
			ENDP
			
			END