/* MacOS.dep.c */
/* 31Dec91  e  */
/* 10Dec92  e  for SML/NJ 92 */
/* 30Dec92  e  fixed "no apname" bug */

/* A MacOS version of Signals and Sbrk for ML */

#include "MacOS.dep.h"
#include "os_mac.h"

#include <errno.h>
#include <stdlib.h>
#include <unix.h>
#include <fcntl.h>
#include <console.h>
#include <string.h>
#include <Packages.h>

/* portions derived from: M68.dep.c
 *
 * COPYRIGHT (c) 1990 by AT&T Bell Laboratories.
 *
 * M68 dependent code for the SML/NJ runtime kernel.
 */

#include "ml_os.h"
#include "ml_state.h"
#include "request.h"

#include "tags.h"
#include "ml_types.h"

#include <sysEqu.h>		/* for CurrentA5 */
#include <Dialogs.h>
#include <Memory.h>
#include <Events.h>
#include <Files.h>
#include <StandardFile.h>
#undef  SystemSixOrLater
#define SystemSixOrLater 1
#include <OSUtils.h>

extern void cecho2file(char *, int, FILE *);
extern long _ftype, _fcreator;
extern int	arenasize;
extern int	isExported;

int mac_sbrk_done = FALSE;
int mac_brkpt = 0;
int mac_blimt = 0;

char usual_prefix[] = ":";
char *prefix = usual_prefix;

jmp_buf restartJmpBuf;

/** Define trap numbers **/

#ifndef _Unimplemented
#define _Unimplemented	0xA89F	/* Unimplemented trap	*/
#endif

#ifndef _SysEnvirons
#define _SysEnvirons	0xA090	/* SysEnvirons Inquiry	*/
#endif

static char *eargv[4];
static char apname[64];

static short dfltWDVRefNum;
static short homeWDVRefNum;
static long  dfltWDDirID;
static long  homeWDDirID;

static void getHome(void)	{
	WDPBRec pb;
	
	pb.ioNamePtr = 0;
	pb.ioCompletion = 0;
	PBHGetVolSync(&pb);
	dfltWDVRefNum = homeWDVRefNum = pb.ioWDVRefNum;
	dfltWDDirID = homeWDDirID = pb.ioWDDirID;
}

static void setupDfltDir(void)	{
	WDPBRec pb;
	
	pb.ioNamePtr = 0;
	pb.ioCompletion = 0;
	pb.ioVRefNum = dfltWDVRefNum;
	pb.ioWDDirID = dfltWDDirID;
	PBHSetVolSync(&pb);
}

static void setupHomeDir(void)	{
	WDPBRec pb;
	
	pb.ioNamePtr = 0;
	pb.ioCompletion = 0;
	pb.ioVRefNum = homeWDVRefNum;
	pb.ioWDDirID = homeWDDirID;
	PBHSetVolSync(&pb);
}

int mac_init( char ***p_argv )
{
    SysEnvRec sysenv;
    KeyMap kMap;
    long len;
    short err, err2;
    short refNum;
    
    /* make sure System 6 or later is running */
    if ( GetTrapAddress(_SysEnvirons) == GetTrapAddress(_Unimplemented) )
    	die("SMLeNJ requires at least System 6");
    /* make sure it's a M68020+ and has an FPU */
    SysEnvirons( curSysEnvVers, &sysenv );
    if ( (sysenv.processor < env68020) || (!sysenv.hasFPU) )
    	die("SMLeNJ requires at least a M68020 and FPU");
    /* added this line for Soren/Quadra...   10Feb92  e  */
    SetApplLimit((void *)(((long )&sysenv)-32768L-2000L));
    /* find the home directory */
    getHome();
    setjmp(restartJmpBuf);
    arenasize = 0;
    isExported = 0;
    prefix = usual_prefix;
    _ftype = 'SMLi';
    _fcreator = 'NJML';
	sprintf(apname, "%#s", CurApName);				/* magic bullet  to prevent prefix */
    thinkc_workaround();   /* see cstruct.c */
    /* err = HOpenDF( homeRefNum, 0, (unsigned char *)CurApName, fsRdPerm, &refNum ); 08Jan93 e */
    err = HOpenDF( dfltWDVRefNum, dfltWDDirID, (unsigned char *)CurApName, fsRdPerm, &refNum );
    if ( err == noErr )
    { err2 = GetEOF( refNum, &len );
      FSClose( refNum );
      if ( err2 == noErr && len > 500000 )
	  { GetKeys( kMap );
		if ( ! ( kMap[1] & 4 ) )
		{ /* option key not down */
		  chatting( "[Loading SML core image]\n");	/* note: NEEDED for console_init() */
		  eargv[0] = apname;
		  eargv[1] = "-i";
		  eargv[2] = apname;
		  eargv[3] = 0;
		  *p_argv = eargv;
		  return 3;
		}
	  }
    }
    return ccommand(p_argv);
}

#if 0
static char *
strcatbut(/* char *s1, const char *s2 */)
{
	asm {
		movea.l	4(sp),a0		;  A0 = s1
		movea.l	8(sp),a1		;  A1 = s2
		clr.l	d0				;  oops, otherwise d0 might never be set - 7Jan92 
@1		tst.b	(a0)+
		bne.s	@1
		subq.l	#1,a0
@2		move.b	(a1)+,d1		;  D1.B = char
		cmpi.b	#'.',d1
		bne.s	@3
		move.l	a1,d0			;  D0.L = addr of last '.'
		subq.l	#1,d0
		cmpi.b	#'.',(a1)
		bne.s	@4
		addq.l	#1,a1
		bra.s	@2
@3		cmpi.b	#'/',d1
		bne.s	@4
		moveq	#':',d1
@4		move.b	d1,(a0)+
		bne.s	@2
	}
}
#else
/* 11Aug93  e */
static char *
strcatbut(/* char *s1, const char *s2 */)
{
	asm {
		movea.l	4(sp),a0		;  A0 = s1
		movea.l	8(sp),a1		;  A1 = s2
		clr.l	d0				;  oops, otherwise d0 might never be set - 7Jan92 
@1		tst.b	(a0)+
		bne.s	@1
		subq.l	#1,a0
@2		move.b	(a1)+,d1		;  D1.B = char
		cmpi.b	#'.',d1
		bne.s	@3				; '.' ?
		cmpi.b	#'.',(a1)		; '..' ?
		bne.s	@5
		addq.l	#1,a1			;  skip '..'
		tst.b	(a1)			;  at end ?
		bne.s	@2				;    if not '..' -> ''
		bra.s	@7				;    if so  '..' -> ':'
@5		cmpi.b	#'/',(a1)		; './' ?
		bne.s	@6
		addq.l	#1,a1			;           './' -> ''
		bra.s	@2
@6		tst.b	(a1)			; ends in '.' ?     // 24Aug93 e
		beq.s	@8				;    if so, drop it // 24Aug93 e
		move.l	a1,d0
		subq.l	#1,d0			;  D0.L = addr of last '.'
		bra.s	@4
@3		cmpi.b	#'/',d1         ; '/' ?
		bne.s	@4
@7		moveq	#':',d1			;            '/' -> ':'
@4		move.b	d1,(a0)+
		bne.s	@2
@8	}
}
#endif

extern unsigned char * __c2p(register const char *s, register char *t);

/* 12Sep93  e */
										/* unix name to mac name */
int e_u2m( char *dn, char *cbuf ) {
	char pbuf[256];

	if((strlen(prefix)+strlen(dn)) > 255) {
		errno = EDOM;
		return -1;
	}
	/* keep this translation consistent with that done in e_open below */
	setupDfltDir();
	if ( strchr( dn, ':' ) )
		strcpy( cbuf, dn );
	else {
		if ( *dn == '' ) {
			setupHomeDir();
      		dn++;
		}
		if ( *dn == '/' ) { *cbuf = 0; dn++; }
		else strcpy( cbuf, prefix );
    	strcatbut( cbuf, dn );
    }
    return 0;
}

extern int e_open_string_stream( char *str );

static int dev_null;

int e_open_dev_null(void)
{		/* thanks, Peter Canning  --  24Aug93  e  */
        return dev_null ? dev_null : (dev_null = e_open_string_stream(""));
}

/* 14Aug93  e */
/* SML/NJ (and UNIX) allow opening a directory, assigning it a fd
   and provide read operations on the directory; I'm attempting to
   mimic that here...
*/
#define qDFDs 4
#define eDFDOFFS 65536

static struct _open_dirs
{	short openp;
	short numFiles;
	short index;
	short ioVRefNum;
	long dirID;
} open_dirs[qDFDs];

int e_close( int fd )
{	int dd;
	struct _open_dirs *dp;

	if ( dev_null && fd == dev_null )
		return 0; /* don't really close it */
	if ( ((unsigned int)(dd = (fd - eDFDOFFS))) >= qDFDs )
		return close(fd); /* not a DFD */
	dp = &open_dirs[dd];
	dp->openp = 0; /* not open */
	return 0;
}

static int e_opendir( char *dn )
{	FSSpec fss;
	char pbuf[256];
	CInfoPBRec pb;
	struct _open_dirs *dp;
	int m,n,o;
	
	if( *dn == 0 ) goto labelMakeRoot;

    errno = FSMakeFSSpec( dfltWDVRefNum, dfltWDDirID, __c2p( dn, pbuf ), &fss );
	if ( errno )
	{	if ( errno == -120 && strspn(dn,":") == strlen(dn) )
		{ /* all ":" => the "root" */
labelMakeRoot:
		  errno = 0;
		  pb.dirInfo.ioDrDirID = -1;
		  fss.vRefNum = -1;
		  pb.dirInfo.ioDrNmFls = 4; /* fix? */
		  o = 2;
		  goto labelDNalloc;
		}
		return -1;
	}
	/* set-up paramter block to get # of files/folders */
	pb.dirInfo.ioVRefNum = fss.vRefNum;
	pb.dirInfo.ioDrDirID = fss.parID;
	pb.dirInfo.ioNamePtr = fss.name;
	pb.dirInfo.ioFDirIndex = 0;	/* use name & vRefNum & dirID */
	
	/* get number of files/folders in dir */
	if ( PBGetCatInfo(&pb,false) )
		return -1;
	if ((pb.dirInfo.ioFlAttrib & 0x10) == 0)
	   return -1; /* not a directory */
	o = 1;

labelDNalloc:
	for ( m = -1, n = 0; n < qDFDs; n++ )
	{	dp = &open_dirs[n];
		if ( dp->openp )
		{	if ( (dp->dirID == pb.dirInfo.ioDrDirID)
				 && (dp->ioVRefNum == fss.vRefNum)
				 && (dp->openp == o) )
				return n + eDFDOFFS;
		}
		else if (m < 0) m = n;
	}
	if (m < 0) return -1; /* no more room */
	dp = &open_dirs[m];
	dp->dirID = pb.dirInfo.ioDrDirID;
	dp->ioVRefNum = fss.vRefNum;
	dp->numFiles = pb.dirInfo.ioDrNmFls;
	dp->index = 1;
	dp->openp = o;
	return m + eDFDOFFS;
}

int e_getdents( int fd, char *buf, int sz )
{	int dd;
	struct _open_dirs *dp;
	Str255 fName;			/* filename storage */
	unsigned char *p = (unsigned char *)buf;	

	if ( ((unsigned int)(dd = (fd - eDFDOFFS))) >= qDFDs )
		return -1; /* not a DFD */
	dp = &open_dirs[dd];
	if ( !dp->openp )
		return -1; /* not open */
	
	if ( dp->openp == 2 )
	{	VolumeParam pb;
		pb.ioCompletion = 0;
		pb.ioNamePtr = fName;
		pb.ioVolIndex = dp->index;
		while ( ! PBGetVInfoSync((ParmBlkPtr)&pb) )
		{	if ( ( (char *)p - buf + pb.ioNamePtr[0] + 2 ) > sz )
				return (char *)p - buf; /* directory too long, done for now */
			*p++ = pb.ioNamePtr[0] + 2;
			sprintf( (char *)p, "%#s", pb.ioNamePtr );
			p += pb.ioNamePtr[0] + 1;
			dp->index++; /* increment volume index */
			pb.ioVolIndex = dp->index;
		}
	}
	else
	{	CInfoPBRec pb;
		pb.dirInfo.ioVRefNum = dp->ioVRefNum;
		pb.dirInfo.ioFDirIndex = 0;	/* use name & vRefNum & dirID */
		pb.dirInfo.ioNamePtr = fName;
	
		while (dp->index <= dp->numFiles)
		{	pb.dirInfo.ioDrDirID = dp->dirID;
			pb.dirInfo.ioFDirIndex = dp->index;
			if (PBGetCatInfo(&pb,false))
				return -1;
			if ( ( (char *)p - buf + pb.dirInfo.ioNamePtr[0] + 2 ) > sz )
				return (char *)p - buf; /* directory too long, done for now */
			*p++ = pb.dirInfo.ioNamePtr[0] + 2;
			sprintf( (char *)p, "%#s", pb.dirInfo.ioNamePtr );
			p += pb.dirInfo.ioNamePtr[0] + 1;
			dp->index++; /* increment file index */
		}
	}
	return (char *)p - buf;	
}

/* 13Aug93  e  -- repaired version of ThinkC fstat */

static int stat_guts( HFileInfo	*hpb, struct stat *sr )
{
	errno = PBGetCatInfoSync((CInfoPBRec *)hpb);
	if (errno)
		return -1;

	// fill in the 'stat' structure
	if(hpb->ioFlAttrib & 16)
	{	sr->st_mode = S_IFDIR;
		sr->st_size = ((CInfoPBRec *)hpb)->dirInfo.ioDrNmFls;	// number of files
	}
	else
	{	sr->st_mode = S_IFREG;
		sr->st_size = hpb->ioFlLgLen;				// logical data fork length
	}
	sr->st_dev = 0; // hpb->ioVRefNum; 				// vRefNum of volume
	sr->st_ito = hpb->ioDirID;						// hard file number
	sr->st_nlink = 0;								// no links
	sr->st_uid = getuid();							// use default uid and
	sr->st_gid = getgid();							// gid
	sr->st_atime = sr->st_mtime = hpb->ioFlMdDat - TimeBaseDif;	// access/mod dates
	sr->st_ctime = hpb->ioFlCrDat - TimeBaseDif;				// creation date
	return 0;
}

int e_fstat(int fd, struct stat *sr)
{
	CInfoPBRec	pb;
	HFileInfo	*hpb = (HFileInfo *)&pb;
	int dd;
	
	if ( ((unsigned int)(dd = (fd - eDFDOFFS))) < qDFDs )
	{	struct _open_dirs *dp = &open_dirs[dd];
		if ( !dp->openp )
			return -1; /* not open */
		if ( dp->openp == 2 )
		{	/* root */
			sr->st_mode = S_IFDIR;
			sr->st_size = 0;							// number of files
			sr->st_dev = 0; // hpb->ioVRefNum; 			// vRefNum of volume
			sr->st_ito = 0;								// 
			sr->st_nlink = 0;							// no links
			sr->st_uid = getuid();						// use default uid and
			sr->st_gid = getgid();						// gid
			sr->st_atime = sr->st_mtime = TimeBaseDif;	// access/mod dates
			sr->st_ctime = TimeBaseDif;					// creation date
			return 0;
		}
		hpb->ioCompletion = 0;
		hpb->ioVRefNum = dp->ioVRefNum;
		hpb->ioFDirIndex = -1;
		hpb->ioNamePtr = 0L;
		hpb->ioDirID = dp->dirID;
	}
	else
	{	FILE *f;
		unsigned char fn[63];
		FCBPBRec fpb;
		f = fdopen(fd,"");
		fpb.ioCompletion = nil;
		fpb.ioFCBIndx = 0;
		fpb.ioRefNum = f->refnum;
		fpb.ioNamePtr = fn;
		errno = PBGetFCBInfoSync(&fpb);
		if (errno)
			return -1;
		// like __stat
		hpb->ioCompletion = 0;
		hpb->ioVRefNum = fpb.ioFCBVRefNum;
		hpb->ioFDirIndex = 0;
		hpb->ioNamePtr = fn;
		hpb->ioDirID = fpb.ioFCBParID;
	}
	return stat_guts( hpb, sr );
}

int e_lstat(char *fn, struct stat *sr)
{	char mn[256];
	char pn[256];
	CInfoPBRec	pb;
	HFileInfo	*hpb = (HFileInfo *)&pb;

	if ( e_u2m( fn, mn ) != 0 )
		return -1;
	// like __stat
	hpb->ioCompletion = 0;
	hpb->ioVRefNum = hpb->ioFDirIndex = 0;
	hpb->ioNamePtr = __c2p(mn,pn);
	hpb->ioDirID = 0L;
	return stat_guts( hpb, sr );
}

static int getfid_root(char *res)
{	strcpy(res,"");	/* root */
	return 0;
}

static int getfid_guts( short ioVRefNum, long ioDirID, unsigned char *pbuf, char *res )
{	FSSpec fss;
	short err;

    err = FSMakeFSSpec( ioVRefNum, ioDirID, pbuf, &fss );
	if ( err == noErr )
	{	sprintf( res, "%04hX|%08lX|%#s", fss.vRefNum, fss.parID, fss.name );
		return 0;
	}
	else if ( err == -120 && pbuf[0] != 0 && pbuf[1] == ':' )
	{	for( err = 2; err <= pbuf[0]; err++ )
			if ( pbuf[err] != ':' )
			{	/* try from the root */
				unsigned char nbuf[256];
				short i;
				err -= 1;
				nbuf[0] = pbuf[0] - err;
				for( i = 1; i <= nbuf[0]; i++ ) nbuf[i] = pbuf[i+err];
				nbuf[i] = ':';
				if ( nbuf[i-1] != ':' ) nbuf[0] += 1;	
				return getfid_guts( 0, 0, nbuf, res );
			}
		return getfid_root(res);
	}
	return -1;
}

int e_fd_getfid( int fd, char *res ) {
	int dd;
	
	if ( ((unsigned int)(dd = (fd - eDFDOFFS))) < qDFDs )
	{	struct _open_dirs *dp = &open_dirs[dd];
		if ( !dp->openp )
			return -1; /* not open */
		if ( dp->openp == 2 )
			return getfid_root(res); /* root */
		return getfid_guts( dp->ioVRefNum, dp->dirID, "\p", res );
	}
	else
	{	FILE *f;
		unsigned char fn[63];
		FCBPBRec fpb;
		f = fdopen(fd,"");
		fpb.ioCompletion = nil;
		fpb.ioFCBIndx = 0;
		fpb.ioRefNum = f->refnum;
		fpb.ioNamePtr = fn;
		errno = PBGetFCBInfoSync(&fpb);
		if (errno)
			return -1;
		return getfid_guts( fpb.ioFCBVRefNum, fpb.ioFCBParID, fn, res );
	}
}

int e_fn_getfid( char *dn, char *res ) {
	char cbuf[256];
	char pbuf[256];
	short ioVRefNum;
	long ioDirID;

	if((strlen(prefix)+strlen(dn)) > 255) {
		errno = EDOM;
		return -1;
	}
	/* keep this translation consistent with that done in e_open below */
	ioVRefNum = dfltWDVRefNum;
	ioDirID = dfltWDDirID;
	if ( strchr( dn, ':' ) ) {
		__c2p( dn, pbuf );
	}
	else {
		if ( *dn == '' ) {
			ioVRefNum = homeWDVRefNum;
			ioDirID = homeWDDirID;
      		dn++;
		}
		if ( *dn == '/' ) { *cbuf = 0; dn++; }
		else strcpy( cbuf, prefix );
    	strcatbut( cbuf, dn );
		__c2p( cbuf, pbuf );
    }
    if ( *pbuf == 0 )
    	return getfid_root(res);
	return getfid_guts( ioVRefNum, ioDirID, (unsigned char *)pbuf, res );
}

int chdir( char *dn ) {
	WDPBRec pb;
	char cbuf[256];
	char pbuf[256];
	short err;

	if((strlen(prefix)+strlen(dn)) > 255) {
		errno = EDOM;
		return -1;
	}
	/* keep this translation consistent with that done in e_open below */
	if ( strchr( dn, ':' ) ) {
		__c2p( dn, pbuf );
	}
	else {
		if ( *dn == '' ) {
			dfltWDVRefNum = homeWDVRefNum;
			dfltWDDirID = homeWDDirID;
      		dn++;
		}
		if ( *dn == '/' ) { *cbuf = 0; dn++; }
		else strcpy( cbuf, prefix );
    	strcatbut( cbuf, dn );
		__c2p( cbuf, pbuf );
    }
	pb.ioNamePtr = (StringPtr )pbuf;
	pb.ioCompletion = 0;
	pb.ioVRefNum = dfltWDVRefNum;
	pb.ioWDDirID = dfltWDDirID;
	err = PBHSetVolSync(&pb);
	if ( err == noErr )
	{	pb.ioNamePtr = 0;
		pb.ioCompletion = 0;
		PBHGetVolSync(&pb);
		{	if ( err == noErr )
			dfltWDVRefNum = pb.ioWDVRefNum;
			dfltWDDirID = pb.ioWDDirID;
			return 0;
		}
	}
	return -1;
}

int e_mkdir( char *dn, int mode ) /* 9Sep93  e */
{
	/* mode not handled yet */
	FSSpec fss;
	int len;
	long dirID;

	len = strlen(dn);
	if( len > 63 || len <= 0 || strchr( dn, ':' ) || strchr( dn, '/' ) ) {
		/* one directory at a time! */
		errno = EDOM;
	}
	else
	{	__c2p( dn, (char *)fss.name );
		fss.vRefNum = dfltWDVRefNum;
		fss.parID   = dfltWDDirID;
		errno = FSpDirCreate( &fss, smSystemScript, &dirID );
	}
	return (errno ? -1 : 0);
}

/* 4Feb93  e  */
OSType e_getfiletype( char *fn )
{	FInfo finf;
	Str255 pn;
	
	__c2p( fn, (char *)pn );
	if ( HGetFInfo( 0, 0L, pn, &finf ) == noErr )
		return finf.fdType;
	else
		return (OSType )'?!?!';
}

char *texttypes[] = { ".sig", ".sml", ".s", ".txt", ".lex", ".grm", ".desc", ".link", "" };

int e_open(char *fn, int mode)
    {
    char buf[256];
    char *p, **q;
    int res;
    
    if(!strcmp(fn,"/dev/null"))
    	return e_open_dev_null();
    
	/* keep this translation consistent with that done in chdir above */
	if((strlen(prefix)+strlen(fn)) > 255) {
		errno = EDOM;
		return -1;
	}
	setupDfltDir();
	if ( strchr(fn,':') ) {
		strcpy(buf, fn);
    	p=strrchr(buf,'.');
	}
	else {
		if ( *fn == '' ) {
			setupHomeDir();
			if(!strcmp(fn,apname)) _ftype = 'APPL';
      		fn++;
		}
		if ( *fn == '/' ) { *buf = 0; fn++; }
		else strcpy(buf, prefix);
    	p = strcatbut(buf,fn);
    }
    if ( ( ( ! p ) || strcmp( p, ".mo" ) ) && e_getfiletype( buf ) == 'TEXT' )
    	 mode |= O_TEXT;
    else if (p) {
    	for ( q = &texttypes[0]; **q != '\0'; q++ )
    		if ( ! strcmp( p, *q ) ) mode |= O_TEXT;
    }
    if (mode & O_TEXT) _ftype = 'TEXT';
    res = open(buf, mode);
    _ftype = 'SMLi';
    if (res < 0) return e_opendir(buf);
    return res;
}

/* 15Aug93  e  --this is based on getfullpath() in os_mac.c 
**     but makes a UNIX pathname from the current directory
**  Path should point to at least 8192 bytes of free memory */

int e_pwd( char *Path, int MaxLength )
{
  char *p, *q;
  int len;
  CInfoPBRec block;
  Str255 directoryName;

  p = &Path[MaxLength-2];
  *--p = '\0';

  block.dirInfo.ioNamePtr = directoryName;
  block.dirInfo.ioDrParID = dfltWDDirID;

  do
  { block.dirInfo.ioVRefNum = dfltWDVRefNum;
    block.dirInfo.ioFDirIndex = -1;
    block.dirInfo.ioDrDirID = block.dirInfo.ioDrParID;

    if (PBGetCatInfo( &block, FALSE ) != noErr) return -1;
    *--p = '/';
    len = (unsigned char)directoryName[0];
    p -= len;
    if (p < Path) goto e_too_long;
    BlockMove( &directoryName[1], p, len );
  } while ( block.dirInfo.ioDrDirID != fsRtDirID );

  q = Path;
  *q++ = '/';
  while (*p != '\0') *q++ = *p++;
  *q++ = '\n';
  *q = '\0';

  return ( q - Path );

e_too_long:
  return -1;
}

int e_full_path( char *fn, char *Path, int MaxLength )
{   int len, sz;
	char buf[256];
	
	if( MaxLength < 255 ) return -1;
	len = e_pwd( buf, 255 );
	if( len >= 0 && (len + strlen(fn)) <= 255 )
	{	buf[len-1] = 0;
		strcat( buf, fn );
		if( !e_u2m( buf, Path ) ) return strlen(Path);
	}
	return -1;
}

int e_date( char *bf, int MaxLength )
{
	unsigned long	time;
	unsigned char   timeString[32];
	unsigned char   dateString[64];
	GetDateTime( &time );
	IUTimeString( time, 0, timeString );
	IUDateString( time, abbrevDate, dateString );
	sprintf( bf, "%#s %#s\n", timeString, dateString );
	return 0;
}

#if 0

static Point putWhere = { 106, 104 };

/* dribble was... [didn't work!?]
			{
			char **av;
			int ac;
			ac = ccommand(&av);
			}
			break;
*/

void doDribble(void)
	{
	SFReply reply;
	char buf[256];
	IOParam pb;
	
	/*  present dialog  */
	SFPutFile(putWhere, (ConstStr255Param)"", (ConstStr255Param)"", (DlgHookProcPtr)NULL, &reply);
	if (reply.good) {
		/*  redirect stdout  */
		/* setfile((char *) scratch.fName, &scratch); */
		setHome(reply.vRefNum);
		/* sprintf(buf, "%#s", reply.fName); */
		cecho2file((char *) reply.fName, 0, stdout);
		setHome(homeRefNum);
	}
}

#endif

void nalert(char *s)
{
	register int i;
	char p0[255];
	p0[0] = (char) strlen(s);
	strcpy(p0+1,s);
	for(i=1;i<255;i++)
		if (p0[i] == '\n') p0[i] = '\r';
		
	InitCursor();		/* get standard arrow cursor */
	ParamText((ConstStr255Param)p0,"\p","\p","\p");
	NoteAlert((short )ok_alertID,0L);
}

void eDoDialog(int sel)
{
	
	switch(sel) {
	  case 0: {
		DialogPtr dlg;
		short i;
		/* do About */
		/* SetDAFont(helvetica); */
		dlg=GetNewDialog(about_dlogID,NULL,(WindowPtr)(-1L));
		ModalDialog((ModalFilterProcPtr)NULL,&i);
		DisposDialog(dlg);
		/* SetDAFont(systemFont); */
		break;
		}
	  case 1: {
		char s[64];
		sprintf(s,"Heap size is %ldk.\n\nPartition size is %ldk.",
		          arenasize/1024L, (mac_blimt - mac_brkpt)/1024L);
		nalert(s);
		break;
		}
#if 0
	  case 2: {
		doDribble();
		break;
		}
#endif
	}
}

/*  Sbrk */
/**
 ** This implements sbrk/brk using MacOS.
 ** it is based on the mach implementation in callgc.c
 ** 01Jan92  e
 **/

int sbrk(incr)
    int incr;
{
    Size largestSz, cangrowSz;
    Ptr mptr;

    if (incr)
		die("sbrk called with nonzero value");
    if (mac_sbrk_done == FALSE) {
		MaxApplZone();	/* expand the heap */
		largestSz = MaxMem(&cangrowSz);
		cangrowSz = largestSz + cangrowSz - 393216L;
		/*384L*1024L=393216=384k for runtime system*/
		if(!(mptr = NewPtr(cangrowSz)))
			die("sbrk cannot allocate: %dK", cangrowSz/1024);
		mac_brkpt = ((int)(mptr + 8191) & -8192); /* make it a "page" boundary */
		mac_blimt =  (int)(mptr + cangrowSz);
		mac_sbrk_done = TRUE;
    }
    return(mac_brkpt);
}

int brk(pos)
    int pos;
{
    if (pos > mac_blimt)
	return TRUE;
    else
	return FALSE;
}

int getpagesize()	{
    return(8192);
}

int get_edata()	{
    return((mac_blimt + 8191) & -8192);
}

int get_etext()	{
    die("get_etext: not done yet!");
    return(0);
}

/* signals */

static __sig_func_e e_sig[__NSIG_e+1];

void
e_raise_guts(struct sigcontext *scp)
{
	__sig_func_e f;
	int i = scp->sig;
	
	if (i <= 0 || i > __NSIG_e) {
		errno = EINVAL;
		return;
	}
	f = e_sig[i];
	if (f != SIG_IGN_e) {
		/* e_sig[i] = SIG_DFL_e; */
		if (f == SIG_DFL_e)
			_exit(0);
		(*f)(i, (scp->code & 0x0fff), scp);
	}
}

void
init_TRAPS();
void
handleTRAP()
	{
	asm {
@handleTRAP:						; rewritten to handle Virtual Memory -- 10Feb92  e
		MOVE.L	#SIGTRAP, -(SP)		; push sig
		MOVEM.L	D0-D7/A0-A6, -(SP)	;  push 15 regs
		MOVE.L	SP, -(SP)			;   push sig context ptr
		MOVE.L	CurrentA5, A5		;    for C variable access
		JSR		e_raise_guts		;    called in supervisor mode!!!!!
		ADDQ.L	#4, SP				;   drop sig context ptr
		MOVEM.L	(SP)+, D0-D7/A0-A6	;  pop  15 regs
		ADDQ.L	#4, SP				; drop sig
		RTE							; return
		;
@handleFPU:							; added to handle M68882 FPU EXC_PEND -- 12Feb92  e
		MOVE.L	#SIGTRAP, -(SP)		; push sig
		MOVEM.L	D0-D7/A0-A6, -(SP)	;  push 15 regs
		MOVE.L	SP, -(SP)			;   push sig context ptr
		MOVE.L	CurrentA5, A5		;    for C variable access
		JSR		e_raise_guts		;    called in supervisor mode!!!!!
		ADDQ.L	#4, SP				;   drop sig context ptr
		FSAVE	-(SP)				; dump the FPU
		; M68881 check not necessary
		; MOVEQ	#$18, D0			; frame size for M68881
		; CMP.B	D0, 1(SP)			; FPU frame size
		; BEQ.S	@rtn				; if M68881, we're done
		MOVE.B	(SP), D0			; null frame?
		BEQ.S	@rtn				; if so, done
		CLR.L	D0					; get frame size
		MOVE.B	1(SP), D0			; into D0
		BSET	#3, (SP, D0)		; set EXC_PEND bit of BIU
@rtn:	FRESTORE (SP)+				; load the FPU
		MOVEM.L	(SP)+, D0-D7/A0-A6	;  pop  15 regs
		ADDQ.L	#4, SP				; drop sig
		RTE							; return
		;
extern init_TRAPS:
		LEA.L	@handleTRAP, A0
		MOVE.L	A0, 0x14			; dbz
		MOVE.L	A0, 0x1C			; TRAPcc
		LEA.L	@handleFPU, A0		;				12Feb92  e
		MOVE.L	A0, 0xC8			; FPdbz
		MOVE.L	A0, 0xD4			; FPovf
		MOVE.L	A0, 0xCC			; FPund
		; for Virtual Memory Mode					10Feb92  e
		; this is not strictly necessary, 
		;   but speeds things up a bit
		MOVEQ	#0, D0				; in VM Mode?
		DC.W	0xA08D				; _DebugUtil
		SUBQ.L	#8, D0				; VM?
		BLT.S	@punt				; skip if not
		MOVEQ	#8, D0				; enter supervisor mode
		DC.W	0xA08D				; _DebugUtil
		LEA.L	@handleTRAP, A0
		MOVEC	VBR, A1
		MOVE.L	A0, 0x14(A1)		; dbz
		MOVE.L	A0, 0x1C(A1)		; TRAPcc
		LEA.L	@handleFPU, A0		;				12Feb92  e
		MOVE.L	A0, 0xC8(A1)		; FPdbz
		MOVE.L	A0, 0xD4(A1)		; FPovf
		MOVE.L	A0, 0xCC(A1)		; FPund
		MOVE	D0, SR				;  exit supervisor mode
@punt:
	}
}

/* 4Sep93  e */

static int vbl_mask;

int sigblock(int mask)
{	
	int rslt = vbl_mask;
	vbl_mask = mask;
	return rslt;
}

void sigsetmask(int mask)
{
	vbl_mask = mask;
}

/* */

void
e_raise(int sig)
	{
	asm {
		/* fix the stack frame to look like an exception frame */
		MOVE.W	#0, -(SP)			; dummy SR
		MOVE.L	sig, -(SP)			; sig  /* sig was: 6(SP)  -- 10Aug93  e  */
		MOVEM.L	D0-D7/A0-A6, -(SP)	; 15 regs
		MOVE.L	SP, -(SP)			;  sig context ptr
		MOVE.L	CurrentA5, A5		; for C variable access
		JSR		e_raise_guts
		ADDQ.L	#4, SP				; drop context ptr
		MOVEM.L	(SP)+, D0-D7/A0-A6	; 15 regs
		ADDQ.L	#6, SP				; drop dummy SR & sig
	}
	if (sig <= __NSIG)
		signal(sig, e_raise);
}

void
e_signal(int i, __sig_func_e f)
{
	
	if (i <= 0 || i > __NSIG_e) {
		errno = EINVAL;
		return;
	}
	e_sig[i] = f;
	if (i <= __NSIG)
		signal(i, e_raise);
}

void
e_restart_handler(int sig)
	{
	signal(sig, e_restart_handler);
	longjmp(restartJmpBuf, 2);
}

/* ********************  10Aug93  e  --  interval timers  ******************** */

static VBLTask et[3];
static Boolean intr_task_installed[3];
static short   intr_task_interval[3];

long itimer_ticks(long t)
{ int i;
  long r, s;
  r = t;
  for( i = 0; i < 3; i++ )
  	if( intr_task_installed[i]
  	    && (s = intr_task_interval[i] - 1) < r
  	    && s >= 0
  	  )
  	  r = s;
  return r;
}

static pascal void intr_task_0(void)
{
	asm {
		/* fix the stack frame to look like an exception frame */
		MOVE.W	#0, -(SP)			; dummy SR
		PEA		SIGALRM				; the signal
		MOVEM.L	D0-D7/A0-A6, -(SP)	; 15 regs
		MOVE.L	SP, -(SP)			;  sig context ptr
		MOVE.L	CurrentA5, A5		; for C variable access
		TST.L	vbl_mask			; masked?
		BEQ.S	@1					; ...
		MOVE.W	#1,et[0].vblCount 	; punt, vbl tasks are masked
		BRA.S	@2;
@1		JSR		e_raise_guts		; do it
		MOVE.W	intr_task_interval[0],et[0].vblCount ; if == 0 task will be disabled
@2		ADDQ.L	#4, SP				; drop context ptr
		MOVEM.L	(SP)+, D0-D7/A0-A6	; 15 regs
		ADDQ.L	#6, SP				; drop dummy SR & sig
	}
}

static pascal void intr_task_1(void)
{
	asm {
		/* fix the stack frame to look like an exception frame */
		MOVE.W	#0, -(SP)			; dummy SR
		PEA		SIGVTALRM			; the signal
		MOVEM.L	D0-D7/A0-A6, -(SP)	; 15 regs
		MOVE.L	SP, -(SP)			;  sig context ptr
		MOVE.L	CurrentA5, A5		; for C variable access
		TST.L	vbl_mask			; masked?
		BEQ.S	@1					; ...
		MOVE.W	#1,et[1].vblCount 	; punt, vbl tasks are masked
		BRA.S	@2;
@1		JSR		e_raise_guts		; do it
		MOVE.W	intr_task_interval[1],et[1].vblCount ; if == 0 task will be disabled
@2		ADDQ.L	#4, SP				; drop context ptr
		MOVEM.L	(SP)+, D0-D7/A0-A6	; 15 regs
		ADDQ.L	#6, SP				; drop dummy SR & sig
	}
}

static pascal void intr_task_2(void)
{
	asm {
		/* fix the stack frame to look like an exception frame */
		MOVE.W	#0, -(SP)			; dummy SR
		PEA		SIGPROF				; the signal
		MOVEM.L	D0-D7/A0-A6, -(SP)	; 15 regs
		MOVE.L	SP, -(SP)			;  sig context ptr
		MOVE.L	CurrentA5, A5		; for C variable access
		TST.L	vbl_mask			; masked?
		BEQ.S	@1					; ...
		MOVE.W	#1,et[2].vblCount 	; punt, vbl tasks are masked
		BRA.S	@2;
@1		JSR		e_raise_guts		; do it
		MOVE.W	intr_task_interval[2],et[2].vblCount ; if == 0 task will be disabled
@2		ADDQ.L	#4, SP				; drop context ptr
		MOVEM.L	(SP)+, D0-D7/A0-A6	; 15 regs
		ADDQ.L	#6, SP				; drop dummy SR & sig
	}
}

int start_intr_task( int which, short repeat_ticks, short init_ticks )
{ 
  intr_task_interval[which] = repeat_ticks;
  if ( intr_task_installed[which] )
  { if (init_ticks == 0)
	{ VRemove( (QElemPtr )&et[which] );
	  intr_task_installed[which] = 0;
	}
  	else
  	  et[which].vblCount = init_ticks;
  	return 0;
  }
  else
  { et[which].qType    = vType;
	et[which].vblAddr  = (VBLProcPtr ) which == 0 ? intr_task_0 :
									   which == 1 ? intr_task_1 : intr_task_2;
	et[which].vblCount = init_ticks;
	et[which].vblPhase = 0;
  	intr_task_installed[which] = (VInstall( (QElemPtr )&et[which] ) == noErr);
	return (intr_task_installed[which] - 1);
  }
}

/* end of MacOS.dep.c */
